from __future__ import annotations

import re
from ast import literal_eval
from configparser import ConfigParser
from dataclasses import asdict, fields
from ipaddress import (
    AddressValueError,
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Interface,
)
from pathlib import Path
from typing import Any, ClassVar, Iterator, Literal
from uuid import uuid4

import aiofiles
import keyring
import keyring.errors
import yaml
from pydantic import EmailStr, field_validator
from pydantic.dataclasses import Field
from pydantic.dataclasses import dataclass as py_dataclass
from pydantic.networks import HttpUrl
from pydantic.types import StringConstraints
from sshpubkeys import InvalidKeyError, SSHKey
from typing_extensions import Annotated


class ConfigParserDict(dict):
    def items(self):
        for k, v in super().items():
            if v.startswith("[") and v.endswith("]"):
                for i in literal_eval(v):
                    yield k, i
            else:
                yield k, v


class SystemdConfigParser(ConfigParser):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._dict = ConfigParserDict

    def optionxform(self, optionstr: str) -> str:
        # this stops the keys being lowercased
        return optionstr


@py_dataclass
class HypervisorConfig:
    url: str
    auth_type: Literal["token", "userpass"]
    credential: str
    keychain: bool = True

    @classmethod
    def from_dict(cls, data: dict) -> HypervisorConfig:
        items: dict[str, str | bool | None] = {}

        for _field in fields(cls):
            items[_field.name] = data.get(_field.name)

        return cls(**items)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, HypervisorConfig):
            return False

        return (
            self.url == other.url
            and self.auth_type == other.auth_type
            and self.credential == other.credential
        )

    def real_credential(self) -> str | None:
        if not self.keychain:
            return self.credential

        stored_cred = keyring.get_password("arcane_mage", self.credential)

        return stored_cred


@py_dataclass
class ArcaneCreatorConfig:
    config_path: ClassVar[Path] = Path().home() / ".fluxnode_creator.yaml"

    hypervisors: list[HypervisorConfig] = Field(default_factory=list)
    use_keyring: bool = True
    default_page: str | None = None

    as_dict = asdict

    @classmethod
    def from_fs(cls) -> ArcaneCreatorConfig:
        try:
            with open(ArcaneCreatorConfig.config_path) as f:
                config_raw: str | None = f.read()
        except FileNotFoundError:
            config_raw = None

        if not config_raw:
            return cls()

        try:
            parsed: dict | None = yaml.safe_load(config_raw)
        except yaml.YAMLError:
            parsed = None

        if not parsed:
            return cls()

        hypervisors = [
            HypervisorConfig.from_dict(x) for x in parsed.get("hypervisors", [])
        ]

        return cls(
            hypervisors,
            parsed.get("use_keyring", True),
            parsed.get("default_page"),
        )

    @property
    def has_config(self) -> bool:
        return bool(self.hypervisors)

    def write(self) -> None:
        with open(ArcaneCreatorConfig.config_path, "w") as f:
            f.write(yaml.dump(self.as_dict()))

    def update_default_page(self, page: str | None) -> None:
        self.default_page = page

        self.write()

    def add_hypervisor(self, hypervisor: HypervisorConfig) -> bool:
        if next(filter(lambda x: x == hypervisor, self.hypervisors), None):
            return True

        self.hypervisors.append(hypervisor)

        if self.use_keyring:
            cred_uuid = str(uuid4())
            try:
                keyring.set_password(
                    "arcane_mage", cred_uuid, hypervisor.credential
                )
            except keyring.errors.PasswordSetError:
                return False

            hypervisor.credential = cred_uuid
        else:
            hypervisor.keychain = False

        self.write()

        return True

    def remove_hypervisor(self, hypervisor: HypervisorConfig) -> None:
        try:
            self.hypervisors.remove(hypervisor)
        except ValueError:
            return

        if hypervisor.keychain:
            try:
                keyring.delete_password("arcane_mage", hypervisor.credential)
            except keyring.errors.PasswordDeleteError:
                pass


@py_dataclass
class Link:
    state: Literal["up", "down"]
    address: str
    name: str
    kind: str | None
    index: int

    @classmethod
    def from_dict(cls, data: dict) -> Link:
        field_set = {f.name for f in fields(cls) if f.init}
        filtered = {k: v for k, v in data.items() if k in field_set}

        return cls(**filtered)

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            props[_field.name] = value

        return props

    @property
    def connected(self) -> bool:
        return self.state == "up"

    @property
    def ethernet(self) -> bool:
        # not sure how this applies for wlan interfaces as I don't have one to test
        return (
            self.kind is None
            and self.name != "lo"
            and bool(self.address)
            and self.address != "00:00:00:00:00:00"
        )


@py_dataclass
class Address:
    address: str
    prefixlen: int
    family: Literal["ipv4", "ipv6"]
    index: int

    @classmethod
    def from_dict(cls, data: dict) -> Address:
        field_set = {f.name for f in fields(cls) if f.init}
        filtered = {k: v for k, v in data.items() if k in field_set}

        filtered["family"] = "ipv4" if filtered["family"] == 2 else "ipv6"

        return cls(**filtered)

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            props[_field.name] = value

        return props

    @property
    def as_ip_interface(self) -> IPv4Interface | IPv6Interface:
        if self.family == "ipv4":
            return IPv4Interface(f"{self.address}/{self.prefixlen}")
        else:
            return IPv4Interface(f"{self.address}/{self.prefixlen}")


@py_dataclass
class Route:
    dst: IPv4Network
    gateway: str | None
    scope: Literal["universe", "link"]
    proto: Literal["static", "kernel", "boot", "dhcp"]
    link: str
    prefsrc: IPv4Address | None

    @classmethod
    def from_dict(cls, data: dict) -> Route:
        field_set = {f.name for f in fields(cls) if f.init}
        filtered = {k: v for k, v in data.items() if k in field_set}

        filtered["dst"] = IPv4Network(filtered["dst"])

        if filtered["prefsrc"]:
            filtered["prefsrc"] = IPv4Address(filtered["prefsrc"])

        return cls(**filtered)

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            if isinstance(value, IPv4Address):
                props[_field.name] = str(value)
            else:
                props[_field.name] = value

        return props

    @property
    def is_default(self) -> bool:
        return self.dst.with_prefixlen == "0.0.0.0/0" and bool(self.gateway)

    def __hash__(self) -> int:
        return hash(self.dst) + hash(self.link) + hash(self.gateway)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Route):
            return False

        return (
            self.dst == other.dst
            and self.link == other.link
            and self.gateway == other.gateway
        )


@py_dataclass
class Identifier:
    type: Literal["system-uuid", "mac-address"]
    value: str

    @classmethod
    def from_dict(cls, data: dict) -> Identifier:
        filtered = {}

        for _field in fields(cls):
            filtered[_field.name] = data.get(_field.name)

        return cls(**filtered)

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            props[_field.name] = value

        return props


@py_dataclass
class AddressConfig:
    default_dns: ClassVar[set[IPv4Address]] = {
        IPv4Address("1.1.1.1"),
        IPv4Address("8.8.8.8"),
    }

    address: IPv4Interface
    gateway: IPv4Address
    dns: set[IPv4Address] = Field(default=default_dns)

    @classmethod
    def from_dict(cls, data: dict) -> AddressConfig:
        try:
            address = IPv4Interface(data.get("address"))
            gateway = IPv4Address(data.get("gateway"))
        except AddressValueError as e:
            raise ValueError(str(e))

        dns_raw = data.get("dns")

        if gateway not in address.network:
            raise ValueError(
                "Gateway must be within the same subnet as address"
            )

        dns = (
            set(IPv4Address(x) for x in dns_raw) if dns_raw else cls.default_dns
        )

        return cls(address, gateway, dns)

    def to_dict(self) -> dict:
        parsed: dict = {}

        for _field in fields(self):
            attr = getattr(self, _field.name)

            if isinstance(attr, set):
                parsed[_field.name] = [str(x) for x in attr]
            else:
                parsed[_field.name] = str(attr)

        return parsed

    def to_systemd_networkd_dict(self) -> dict:
        formatted = {
            "Address": str(self.address),
            "Gateway": str(self.gateway),
            "DNS": [str(x) for x in self.dns],
        }

        return formatted


@py_dataclass
class NetworkConfig:
    ip_allocation: Literal["dhcp", "static"] = "dhcp"
    address_config: AddressConfig | None = None
    vlan: int | None = Field(default=None, lt=4095, gt=0)
    rate_limit: Literal[35, 75, 135, 250] | None = None

    @classmethod
    def from_dict(cls, data: dict) -> NetworkConfig:
        ip_allocation: str = data.get("ip_allocation", "")
        vlan: int | None = data.get("vlan")
        rate_limit: int | None = data.get("rate_limit")

        if ip_allocation not in ("dhcp", "static"):
            raise ValueError(f"Invalid ip allocation: {ip_allocation}")
        elif ip_allocation == "dhcp":
            return NetworkConfig(rate_limit=rate_limit)

        address_config_raw = data.get("address_config")

        if not address_config_raw:
            raise ValueError(
                "Network config missing Address Config and static selected"
            )

        address_config = AddressConfig.from_dict(address_config_raw)

        return cls("static", address_config, vlan, rate_limit)

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            if hasattr(value, "to_dict"):
                props[_field.name] = value.to_dict()
            else:
                props[_field.name] = value

        return props

    def systemd_ini_configs(
        self, interface_name: str
    ) -> list[tuple[str, SystemdConfigParser]]:
        confs: list[tuple[str, SystemdConfigParser]] = []

        int_conf = SystemdConfigParser()

        confs.append((f"20-{interface_name}.network", int_conf))

        dhcp = {"DHCP": "yes"}
        static = {"DHCP": "no"}

        address_config = (
            self.address_config.to_systemd_networkd_dict()
            if self.address_config
            else {}
        )

        network_config = static if self.ip_allocation == "static" else dhcp
        network_config |= address_config

        int_conf["Match"] = {"Name": interface_name}

        if self.vlan:
            vlan_interface_name = f"{interface_name}.{self.vlan}"

            vlan_netdev_conf = SystemdConfigParser()
            vlan_int_conf = SystemdConfigParser()

            vlan_netdev_conf["NetDev"] = {
                "Name": vlan_interface_name,
                "Kind": "vlan",
            }
            vlan_netdev_conf["VLAN"] = {"Id": str(self.vlan)}
            vlan_int_conf["Network"] = network_config
            int_conf["Network"] = {"DHCP": "no", "VLAN": vlan_interface_name}

            confs.append((f"20-{vlan_interface_name}.netdev", vlan_netdev_conf))
            confs.append((f"20-{vlan_interface_name}.network", vlan_int_conf))
        else:
            int_conf["Network"] = network_config

        return confs


@py_dataclass
class KeyboardConfig:
    # narrow these
    layout: str = "us"
    variant: str = ""

    @classmethod
    def from_dict(cls, data: dict) -> KeyboardConfig:
        filtered = {}

        for _field in fields(cls):
            value = data.get(_field.name)
            if not isinstance(value, str):
                continue

            filtered[_field.name] = value

        return cls(**filtered)

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            props[_field.name] = value

        return props


@py_dataclass
class SystemConfig:
    hostname: str = Field(min_length=2, max_length=253)
    hashed_console: str = "!"  # No password login
    ssh_pubkey: str | None = None
    keyboard: KeyboardConfig = Field(default_factory=KeyboardConfig)

    @field_validator("ssh_pubkey", mode="before")
    @classmethod
    def validate_ssh_pubkey(cls, value: str | None) -> str | None:
        if not value:
            return value

        key = SSHKey(value, strict=True)

        try:
            key.parse()
        except InvalidKeyError:
            raise ValueError("A public key in OpenSSH format is required")

        return value

    @classmethod
    def from_dict(cls, data: dict) -> SystemConfig:
        hostname = data.get("hostname")
        console = data.get("hashed_console", "!")
        keyboard_raw = data.get("keyboard")
        ssh_pubkey = data.get("ssh_pubkey")

        if not hostname:
            raise ValueError("System config is missing hostname")

        if not console:
            raise ValueError("System config is missing hashed console password")

        keyboard = (
            KeyboardConfig.from_dict(keyboard_raw)
            if keyboard_raw
            else KeyboardConfig()
        )

        return cls(
            hostname,
            console,
            ssh_pubkey,
            keyboard,
        )

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            if hasattr(value, "to_dict"):
                props[_field.name] = value.to_dict()
            else:
                props[_field.name] = value

        return props


@py_dataclass
class Identity:
    ui_property_map: ClassVar[dict] = {
        "flux_id": "fluxId",
        "identity_key": "identityKey",
        "tx_id": "txId",
        "output_id": "outputId",
    }

    flux_id: str
    identity_key: str
    tx_id: str
    output_id: int

    asdict = asdict

    @field_validator("flux_id", mode="after")
    @classmethod
    def validate_flux_id(cls, value: str) -> str:
        id_len = len(value)

        if id_len > 72 or id_len < 14:
            raise ValueError("FluxId must be between 14 and 72 characters")

        return value

    @field_validator("identity_key", mode="after")
    @classmethod
    def validate_identity_key(cls, value: str) -> str:
        key_len = len(value)

        if key_len < 51 or key_len > 52:
            raise ValueError("Identity key must be 51 or 52 characters")

        return value

    @field_validator("tx_id", mode="after")
    @classmethod
    def validate_txid(cls, value: str) -> str:
        if len(value) != 64:
            raise ValueError("Transaction Id must be 64 characters")

        return value

    @field_validator("output_id", mode="before")
    @classmethod
    def validate_output_id(cls, value: str | int) -> int:
        value = int(value)

        if value < 0 or value > 999:
            raise ValueError("OutputId must be between 0 and 999")

        return value

    @classmethod
    def from_dict(cls, data: dict, ui: bool = False) -> Identity:
        """Generates a fluxnode identity from either a UI dict, or standard dict

        Args:
            params (dict): The raw dict to convert to an Identity
            ui (bool, optional): If params are from the UI. Defaults to False.

        Raises:
            ValueError: If the dict doesn't pass validation

        Returns:
            Identity: The fluxnode Identity
        """

        items = []

        for _field in fields(cls):
            name = cls.ui_property_map[_field.name] if ui else _field.name
            prop = data.get(name, None)

            if prop is None:
                raise ValueError(f"Property: {name} missing")

            items.append(prop)

        return cls(*items)

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            props[_field.name] = value

        return props

    @property
    def ui_dict(self) -> dict:
        """Returns the Identity in dict format for the UI

        Returns:
            dict: UI formatted dictionariy
        """

        return {
            self.ui_property_map[_field.name]: getattr(self, _field.name)
            for _field in fields(self)
        }


@py_dataclass
class DiscordNotification:
    ui_property_map: ClassVar[dict] = {
        "webhook_url": "discordWebhookUrl",
        "user_id": "discordUserId",
    }

    watchdog_property_map: ClassVar[dict] = {
        "webhook_url": "web_hook_url",
        "user_id": "ping",
    }

    webhook_url: str | None = None
    user_id: str | None = None

    @field_validator("webhook_url", mode="after")
    @classmethod
    def validate_webhook_url(cls, value: str | None) -> str | None:
        if not value:
            return value

        # this will raise Validation error (and be caught)
        url = HttpUrl(value)
        # discordapp.com is the deprecated endpoint
        valid_hosts = ["discordapp.com", "discord.com"]

        if url.host not in valid_hosts:
            raise ValueError(
                "Discord webhook url must have discord as the host"
            )

        if not url.scheme == "https":
            raise ValueError("discord webhook url scheme must be https")

        if not url.path or not url.path.startswith("/api/webhooks"):
            raise ValueError(
                "discord webhook path must start with /api/webhooks"
            )

        return value

    @field_validator("user_id", mode="before")
    @classmethod
    def validate_user_id(cls, value: str | int) -> str:
        if not value and isinstance(value, str):
            return value
        elif value is None:
            return value

        as_str = str(value)

        len_user_id = len(as_str)

        if len_user_id < 17 or len_user_id > 19:
            raise ValueError(
                "Discord user id must be between 17 and 19 characters"
            )

        return as_str

    @classmethod
    def from_dict(cls, data: dict, ui: bool = False) -> DiscordNotification:
        items = []

        for _field in fields(cls):
            name = cls.ui_property_map[_field.name] if ui else _field.name
            prop = data.get(name, None)

            items.append(prop)

        return cls(*items)

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            props[_field.name] = value

        return props

    @property
    def watchdog_dict(self) -> dict:
        return {
            self.watchdog_property_map[field.name]: getattr(self, field.name)
            or "0"
            for field in fields(self)
        }

    @property
    def ui_dict(self) -> dict:
        return {
            self.ui_property_map[field.name]: getattr(self, field.name)
            for field in fields(self)
        }


@py_dataclass
class TelegramNotification:
    ui_property_map: ClassVar[dict] = {
        "bot_token": "telegramBotToken",
        "chat_id": "telegramChatId",
    }

    watchdog_property_map: ClassVar[dict] = {
        "bot_token": "telegram_bot_token",
        "chat_id": "telegram_chat_id",
        "telegram_alert": "telegram_alert",
    }

    bot_token: str | None = Field(
        None, pattern=r"^[0-9]{8,10}:[a-zA-Z0-9_-]{35}$"
    )
    chat_id: str | None = Field(None, min_length=6, max_length=1000)

    @classmethod
    def from_dict(cls, data: dict, ui: bool = False) -> TelegramNotification:
        items = []

        for _field in fields(cls):
            name = cls.ui_property_map[_field.name] if ui else _field.name
            prop = data.get(name, None)

            items.append(prop)

        return cls(*items)

    @property
    def telegram_alert(self) -> str:
        return "1" if self.bot_token and self.chat_id else "0"

    @property
    def watchdog_dict(self) -> dict:
        # we use the watchdog_property_map so we can include telegram_alert
        return {
            self.watchdog_property_map[key]: getattr(self, key) or "0"
            for key in self.watchdog_property_map
        }

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            props[_field.name] = value

        return props

    @property
    def ui_dict(self) -> dict:
        # ToDo: figure out how to use InitVar with telegram_alert
        return {
            self.ui_property_map[field.name]: getattr(self, field.name)
            for field in fields(self)
        }


@py_dataclass
class Notifications:
    ui_property_map: ClassVar[dict] = {
        "email": "emailAddress",
        "webhook": "genericWebhookUrl",
        "node_name": "nodeName",
    }

    discord: DiscordNotification = Field(default_factory=DiscordNotification)
    telegram: TelegramNotification = Field(default_factory=TelegramNotification)
    email: (
        Annotated[
            EmailStr, StringConstraints(strip_whitespace=True, to_lower=True)
        ]
        | None
    ) = None
    webhook: str | None = None
    node_name: str | None = None

    asdict = asdict

    @field_validator("webhook", mode="after")
    @classmethod
    def validate_webhook(cls, value: str | None) -> str | None:
        if not value:
            return value

        # this will raise ValidationError for us
        HttpUrl(value)

        return value

    @classmethod
    def from_dict(cls, data: dict, ui: bool = False):
        discord_raw = data.get("discord")
        telegram_raw = data.get("telegram")

        discord = (
            DiscordNotification.from_dict(discord_raw, ui)
            if discord_raw
            else DiscordNotification()
        )

        telegram = (
            TelegramNotification.from_dict(telegram_raw, ui)
            if telegram_raw
            else TelegramNotification()
        )

        other_items = (
            {k: data.get(v) for k, v in cls.ui_property_map.items()}
            if ui
            else {k: data.get(k) for k in cls.ui_property_map.keys()}
        )

        return cls(discord=discord, telegram=telegram, **other_items)

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            if hasattr(value, "to_dict"):
                props[_field.name] = value.to_dict()
            elif value:
                props[_field.name] = value

        return props

    def to_ui_dict(self) -> dict:
        discord = self.discord.ui_dict if self.discord else {}
        telegram = self.telegram.ui_dict if self.telegram else {}

        rest = {
            self.ui_property_map[field.name]: getattr(self, field.name)
            for field in fields(self)
            if field.name in self.ui_property_map and getattr(self, field.name)
        }

        return discord | telegram | rest


@py_dataclass
class GravityConfig:
    ui_property_map: ClassVar[dict] = {
        "blocked_ports": "blockedPorts",
        "blocked_repositories": "blockedRepositories",
    }

    debug: bool = False
    development: bool = False
    testnet: bool = False
    blocked_ports: list[int] = Field(default_factory=list)
    blocked_repositories: list[str] = Field(default_factory=list)

    asdict

    @field_validator("blocked_ports", mode="before")
    @classmethod
    def validate_blocked_ports(cls, value: list[int | str]) -> list[int]:
        # ToDo: figure how to do this as a field

        as_int = []

        pattern = r"^([1-9][0-9]{0,3}|[1-5][0-9]{4}|6[0-4][0-9]{3}|65[0-4][0-9]{2}|655[0-2][0-9]|6553[0-5])$"

        for port in value:
            if not re.search(pattern, str(port)):
                raise ValueError(f"Port: {port} must be in the range 0-65535")

            as_int.append(int(port))

        return as_int

    @field_validator("blocked_repositories", mode="before")
    @classmethod
    def validate_blocked_repositories(cls, value: list) -> list:
        # ToDo: figure how to do this as a field

        pattern = r"^(?:(?:(?:(?:[\w-]+(?:\.[\w-]+)+)(?::\d+)?)|[\w]+:\d+)\/)?\/?(?:(?:(?:[a-z0-9]+(?:(?:[._]|__|[-]*)[a-z0-9]+)*)\/){0,2})(?:[a-z0-9-_.]+\/{0,1}[a-z0-9-_.]+)[:]?(?:[\w][\w.-]{0,127})?$"

        for repo in value:
            if not re.search(pattern, repo):
                raise ValueError(f"Repository: {repo} must be a valid format")

        return value

    @classmethod
    def from_dict(cls, data: dict, ui: bool = False) -> GravityConfig:
        items: dict = {}

        for prop, ui_prop in cls.ui_property_map.items():
            key = ui_prop if ui else prop
            value = data.get(key, None)

            if value is not None:
                items[prop] = value

        return cls(**items)

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            if value:
                props[_field.name] = value

        return props

    def to_ui_dict(self) -> dict:
        return {
            self.ui_property_map[field.name]: getattr(self, field.name)
            for field in fields(self)
            if field.name in self.ui_property_map
        }


@py_dataclass
class FluxnodeNetworkConfig:
    upnp_port: int | None = None
    router_address: str | None = None
    private_chain_sources: list[str] = Field(default_factory=list)

    @classmethod
    def from_dict(cls, data: dict) -> FluxnodeNetworkConfig:
        props = {}
        for _field in fields(cls):
            value = data.get(_field.name, None)

            if value is not None:
                props[_field.name] = value

        return cls(**props)

    @property
    def upnp_enabled(self) -> bool:
        return bool(self.upnp_port)

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            if value:
                props[_field.name] = value

        return props

    async def write_installer_config(self, file_path: Path) -> bool:
        # we let the configure app set the local chain sources
        ip_pattern = r"^((25[0-5]|(2[0-4]|1\d|[1-9]|)\d)\.?\b){4}$"
        port_pattern = r"^(?:6553[0-5]|655[0-2][0-9]|65[0-4][0-9]{2}|6[0-4][0-9]{3}|[1-5][0-9]{4}|[1-9][0-9]{0,3})(?:\s?,\s?(6553[0-5]|655[0-2][0-9]|65[0-4][0-9]{2}|6[0-4][0-9]{3}|[1-5][0-9]{4}|[1-9][0-9]{0,3}))*$"

        filtered_chain_sources = []

        for chain_source in self.private_chain_sources:
            try:
                ip, port = chain_source.split(":")
            except ValueError:
                continue

            if not re.match(ip_pattern, ip) or not re.match(port_pattern, port):
                continue

            if not IPv4Address(ip).is_private:
                continue

            filtered_chain_sources.append(chain_source)

        config = {
            "network": {
                "upnp_enabled": self.upnp_enabled,
                "upnp_port": self.upnp_port,
                "private_chain_sources": filtered_chain_sources,
                "router_address": self.router_address,
            }
        }

        writeable_config = yaml.dump(
            config,
            sort_keys=False,
            default_flow_style=False,
        )

        async with aiofiles.open(file_path, "w") as f:
            await f.write(writeable_config)

        return True


@py_dataclass
class FluxnodeConfig:
    config_path: ClassVar[Path] = Path("/mnt/root/config/flux_user_config.yaml")

    identity: Identity
    gravity: GravityConfig = Field(default_factory=GravityConfig)
    network: FluxnodeNetworkConfig = Field(
        default_factory=FluxnodeNetworkConfig
    )
    notifications: Notifications = Field(default_factory=Notifications)

    @staticmethod
    def remove_none_factory(data: list[tuple[str, Any]]):
        return {field: value for field, value in data if value is not None}

    @classmethod
    def from_dict(cls, params: dict) -> FluxnodeConfig:
        identity_raw: dict | None = params.get("identity")
        notifications_raw: dict | None = params.get("notifications")
        gravity_raw: dict | None = params.get("gravity")
        network_raw: dict | None = params.get("network")

        if not identity_raw:
            raise ValueError("Fluxnode identity missing")

        identity = Identity.from_dict(identity_raw)
        notifications = (
            Notifications.from_dict(notifications_raw)
            if notifications_raw
            else Notifications()
        )
        gravity = (
            GravityConfig.from_dict(gravity_raw)
            if gravity_raw
            else GravityConfig()
        )
        network = (
            FluxnodeNetworkConfig.from_dict(network_raw)
            if network_raw
            else FluxnodeNetworkConfig()
        )

        return cls(identity, gravity, network, notifications)

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            if hasattr(value, "to_dict"):
                props[_field.name] = value.to_dict()
            elif value:
                props[_field.name] = value

        return props

    @classmethod
    async def from_config_file(cls) -> FluxnodeConfig | None:
        try:
            async with aiofiles.open(cls.config_path, "r") as f:
                data = await f.read()
        except FileNotFoundError:
            return None

        try:
            conf: dict = yaml.safe_load(data)
        except yaml.YAMLError:
            return None

        return cls.from_dict(conf)

    async def write_user_config(
        self, file_path: Path, ssh_pubkey: str | None = None
    ) -> bool:
        config = {
            "identity": self.identity.to_dict(),
            "notifications": self.notifications.to_dict(),
            "miscellaneous": {"ssh_pubkey": ssh_pubkey}
            | self.gravity.to_dict(),
        }

        writeable_config = yaml.dump(
            config,
            sort_keys=False,
            default_flow_style=False,
        )

        async with aiofiles.open(file_path, "w") as f:
            await f.write(writeable_config)

        return True

    @property
    def fluxd_properties(self) -> dict:
        identity = self.identity

        return {
            "zelnodeprivkey": identity.identity_key,
            "zelnodeoutpoint": identity.tx_id,
            "zelnodeindex": identity.output_id,
        }

    @property
    def gravity_properties(self) -> dict:
        return {
            "flux_id": self.identity.flux_id,
            "debug": self.gravity.debug if self.gravity else False,
            "development": self.gravity.development if self.gravity else False,
            "testnet": self.gravity.testnet if self.gravity else False,
            "blocked_ports": self.gravity.blocked_ports if self.gravity else [],
            "blocked_repositories": self.gravity.blocked_repositories
            if self.gravity
            else [],
        }


@py_dataclass
class Hypervisor:
    node: str
    vm_name: str
    node_tier: Literal["cumulus", "nimbus", "stratus"]
    network: str
    iso_name: str = Field(pattern=r"^FluxLive-\d{10}\.iso$")
    storage_images: str = "local-lvm"
    storage_iso: str = "local"
    storage_import: str = "local"
    vm_id: int | None = None
    startup_config: str | None = None
    disk_limit: int | None = None
    cpu_limit: float | None = None
    network_limit: int | None = None
    start_on_creation: bool = False

    @classmethod
    def from_dict(cls, data: dict) -> Hypervisor:
        props = {}
        for _field in fields(cls):
            value = data.get(_field.name, None)

            if value is not None:
                props[_field.name] = value

        return cls(**props)

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            props[_field.name] = value

        return props


@py_dataclass
class InstallerConfig:
    auto_reboot: bool = True
    reboot_to_firmware: bool = False
    reboot_to_boot_menu: bool = False

    @classmethod
    def from_dict(cls, data: dict) -> InstallerConfig:
        props = {}
        for _field in fields(cls):
            value = data.get(_field.name, None)

            if value is not None:
                props[_field.name] = value

        return cls(**props)

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            props[_field.name] = value

        return props


@py_dataclass
class MetricsAppConfig:
    poweroff_screen: int = 0
    theme: str = "flexoki"

    @classmethod
    def from_dict(cls, data: dict) -> MetricsAppConfig:
        props = {}
        for _field in fields(cls):
            value = data.get(_field.name, None)

            if value is not None:
                props[_field.name] = value

        return cls(**props)

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name)

            props[_field.name] = value

        return props

    async def write_config(self, file_path: Path) -> bool:
        config = self.to_dict()

        writeable_config = yaml.dump(
            config,
            sort_keys=False,
            default_flow_style=False,
        )

        async with aiofiles.open(file_path, "w") as f:
            await f.write(writeable_config)

        return True


@py_dataclass
class ArcaneOsConfig:
    fluxnode: FluxnodeConfig
    system: SystemConfig
    network: NetworkConfig = Field(default_factory=NetworkConfig)
    installer: InstallerConfig = Field(default_factory=InstallerConfig)
    metrics_app: MetricsAppConfig = Field(default_factory=MetricsAppConfig)
    hypervisor: Hypervisor | None = None
    identifier: Identifier | None = None

    @classmethod
    def from_dict(cls, data: dict) -> ArcaneOsConfig:
        fluxnode_raw = data.get("fluxnode")
        system_raw = data.get("system")
        network_raw = data.get("network")
        hypervisor_raw = data.get("hypervisor")
        identifier_raw = data.get("identifier")
        installer_raw = data.get("installer")
        metrics_app_raw = data.get("metrics_app")

        if not fluxnode_raw:
            raise ValueError("fluxnode config missing")

        if not system_raw:
            raise ValueError("system config missing")

        fluxnode = FluxnodeConfig.from_dict(fluxnode_raw)
        system = SystemConfig.from_dict(system_raw)

        network = (
            NetworkConfig.from_dict(network_raw)
            if network_raw
            else NetworkConfig()
        )

        hypervisor = (
            Hypervisor.from_dict(hypervisor_raw) if hypervisor_raw else None
        )

        identifier = (
            Identifier.from_dict(identifier_raw) if identifier_raw else None
        )

        installer = (
            InstallerConfig.from_dict(installer_raw)
            if installer_raw
            else InstallerConfig()
        )

        metrics_app = (
            MetricsAppConfig.from_dict(metrics_app_raw)
            if metrics_app_raw
            else MetricsAppConfig()
        )

        return cls(
            fluxnode,
            system,
            network,
            installer,
            metrics_app,
            hypervisor,
            identifier,
        )

    def to_dict(self) -> dict:
        props = {}

        for _field in fields(self):
            value = getattr(self, _field.name, None)

            if value is not None:
                props[_field.name] = value.to_dict()

        return props

    def as_row(self) -> list[str]:
        address = (
            "dhcp"
            if not self.network.address_config
            else str(self.network.address_config.address)
        )
        return [
            self.hypervisor.node,
            self.system.hostname,
            self.hypervisor.node_tier,
            self.hypervisor.network,
            address,
        ]

    async def write_installer_config(self, file_path: Path) -> bool:
        res = await self.fluxnode.network.write_installer_config(file_path)

        return res

    async def write_user_config(self, file_path: Path) -> bool:
        ssh_pubkey = self.system.ssh_pubkey

        # It makes more sense to move the pubkey to global system config, not as
        # as misc item on the fluxnode config. Will eventually move it off here
        res = await self.fluxnode.write_user_config(
            file_path, ssh_pubkey=ssh_pubkey
        )

        return res

    async def write_metrics_config(self, file_path: Path) -> bool:
        res = await self.metrics_app.write_config(file_path)

        return res


@py_dataclass
class ArcaneOsConfigGroup:
    default_path: ClassVar[Path] = Path("fluxnodes.yaml")

    nodes: list[ArcaneOsConfig] = Field(default_factory=list)

    @classmethod
    def from_dict(cls, data: dict) -> ArcaneOsConfigGroup:
        nodes: list[dict] | None = data.get("nodes")

        if not nodes:
            raise ValueError("ArcaneOsConfigGroup must contain nodes key")

        parsed = [ArcaneOsConfig.from_dict(x) for x in nodes]

        return cls(parsed)

    @classmethod
    def from_fs(cls, config_path: Path | None = None) -> ArcaneOsConfigGroup:
        file_path = config_path or ArcaneOsConfigGroup.default_path

        try:
            with open(file_path, "r") as f:
                data = f.read()
        except (FileNotFoundError, PermissionError):
            # log
            return cls()

        try:
            config_raw = yaml.safe_load(data)
        except yaml.YAMLError:
            # log
            return cls()

        return cls.from_dict(config_raw)

    @property
    def first(self) -> ArcaneOsConfig | None:
        return self.nodes[0] if self.nodes else None

    @property
    def rest(self) -> list[ArcaneOsConfig]:
        return self.nodes[1:] if self.nodes else []

    @property
    def last(self) -> ArcaneOsConfig | None:
        return self.nodes[-1] if self.nodes else None

    def __len__(self) -> int:
        return len(self.nodes)

    def __iter__(self) -> Iterator[ArcaneOsConfig]:
        yield from self.nodes

    def to_dict(self) -> dict:
        return {"nodes": [x.to_dict() for x in self.nodes]}

    def get_node_by_vm_name(
        self, node_name: str, vm_name: str
    ) -> ArcaneOsConfig | None:
        node = next(
            filter(
                lambda x: x.hypervisor
                and x.hypervisor.vm_name == vm_name
                and x.hypervisor.node == node_name,
                self.nodes,
            ),
            None,
        )

        return node

    def get_nodes_by_hypervisor_name(
        self, hyper_name: str
    ) -> ArcaneOsConfigGroup:
        return ArcaneOsConfigGroup(
            list(
                filter(
                    lambda x: x.hypervisor and x.hypervisor.node == hyper_name,
                    self.nodes,
                )
            )
        )

    def add_nodes(self, other: ArcaneOsConfigGroup) -> None:
        self.nodes.extend(other.nodes)


if __name__ == "__main__":
    import asyncio
    from rich.pretty import pprint
    # address_config: AddressConfig | None = None
    # conf = NetworkConfig(
    #     ip_allocation="static",
    #     address_config=AddressConfig(
    #         address=IPv4Interface("192.168.44.14/24"),
    #         gateway=IPv4Address("192.16.44.1"),
    #         dns=["1.1.1.1", "5.5.5.5"],
    #     ),
    # )

    # configs: list[tuple[str, SystemdConfigParser]] = conf.systemd_ini_configs(
    #     "ens44"
    # )

    # for filename, config in configs:
    #     with open(filename, "w") as fd:
    #         config.write(fd, space_around_delimiters=False)

    async def main():
        config = ArcaneOsConfigGroup.from_fs(Path("dev/minimal.yaml"))

        pprint(config)

        await config.nodes[0].write_user_config("testing123.yaml")

    asyncio.run(main())



from __future__ import annotations

from typing import Literal

from textual import work
from textual.app import ComposeResult
from textual.containers import Container, Grid, Horizontal, Vertical
from textual.dom import NoMatches
from textual.reactive import var
from textual.screen import ModalScreen
from textual.validation import URL
from textual.widgets import Button, Footer, Input, Label, Select, Switch

from ..models import HypervisorConfig
from ..proxmox import ProxmoxApi


class AddHypervisorScreen(ModalScreen[HypervisorConfig | None]):
    url_valid = var(False)
    creds_valid = var(False)
    form_valid = var(False)
    hypervisor_valid = var(False)

    has_message = False

    url = var("")
    creds = var("")

    auth_type: var[Literal["token", "userpass"]] = var("token")

    def __init__(self, use_keyring: bool) -> None:
        super().__init__()

        self.use_keyring = use_keyring

    def on_screen_resume(self):
        self.app.set_focus(None)

    def compose(self) -> ComposeResult:
        container = Container()
        container.border_title = "Add Hypervisor"

        info_label = Label("", id="info-label")
        info_label.visible = False

        with container:
            with Grid():
                yield Label("Auth Type:", classes="text-label")
                yield Select(
                    options=[("Token", "token"), ("User / Pass", "userpass")],
                    value="token",
                    allow_blank=False,
                    id="auth-type",
                )
                yield Label("API URL:", classes="text-label")
                yield Input(
                    placeholder="https://your.server:8006",
                    validators=[URL()],
                    id="url-input",
                )
                yield Label("API Token:", classes="text-label", id="auth-label")
                yield Input(
                    placeholder="USER@REALM!TOKENID=UUID",
                    password=True,
                    id="auth-input",
                )
                with Horizontal():
                    yield Button("Reveal", id="reveal")
                    yield Label("Store in Keychain:", classes="text-label")
                    yield Switch(value=self.use_keyring)
            with Vertical():
                yield info_label
                with Horizontal():
                    yield Button("Cancel", id="cancel")
                    yield Button("Validate", id="validate", disabled=True)
                    yield Button("Save", id="save", disabled=True)

        yield Footer()

    def on_input_changed(self, event: Input.Changed) -> None:
        if event.input.id == "url-input":
            self.url = event.value
        elif event.input.id == "auth-input":
            self.creds = event.value

        if self.has_message:
            self.clear_message()

    def compute_url_valid(self) -> bool:
        try:
            valid = self.query_one("#url-input", Input).is_valid
        except NoMatches:
            return False

        return bool(self.url) and valid

    def compute_creds_valid(self) -> bool:
        try:
            valid = self.query_one("#auth-input", Input).is_valid
        except NoMatches:
            return False

        return bool(self.creds) and valid

    def compute_form_valid(self) -> bool:
        return self.url_valid and self.creds_valid

    def watch_form_valid(self, old: bool, new: bool) -> None:
        if old == new:
            return

        try:
            button = self.query_one("#validate", Button)
        except NoMatches:
            return

        button.disabled = not self.form_valid

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "validate":
            self.validate_hypervisor()
        elif event.button.id == "save":
            self.dismiss(HypervisorConfig(self.url, self.auth_type, self.creds))
        elif event.button.id == "cancel":
            self.dismiss(None)
        elif event.button.id == "reveal":
            input = self.query_one("#auth-input", Input)
            input.password = not input.password

    @work(name="validate_hypervisor")
    async def validate_hypervisor(self) -> None:
        client: ProxmoxApi | None = None

        if self.auth_type == "token":
            token = ProxmoxApi.parse_token(self.creds)

            if token:
                client = ProxmoxApi.from_token(self.url, *token)
            else:
                self.set_message(
                    msg_type="error", value="Unable to parse token"
                )

        elif self.auth_type == "userpass":
            user_pass = ProxmoxApi.parse_user_pass(self.creds)

            if user_pass:
                client = await ProxmoxApi.from_user_pass(self.url, *user_pass)
            else:
                self.set_message(
                    msg_type="error", value="Unable to parse user / pass"
                )

        if not client:
            self.hypervisor_valid = False
            return

        async with client.session():
            res = await client.get_hypervisor_nodes()

        if res.unauthorized:
            self.set_message(msg_type="error", value="Unauthorized")
            self.hypervisor_valid = False
        elif res.error:
            self.set_message(msg_type="error", value=res.error)
            self.hypervisor_valid = False
        else:
            self.set_message(msg_type="info", value="Validated")
            self.hypervisor_valid = True

    def clear_message(self) -> None:
        try:
            label = self.query_one("#info-label", Label)
        except NoMatches:
            return

        label.visible = False
        label.update("")
        label.set_class(False, "--info-error")
        label.set_class(False, "--info-success")

        self.has_message = False

    def set_message(
        self, msg_type: Literal["error", "info"], value: str
    ) -> None:
        try:
            label = self.query_one("#info-label", Label)
        except NoMatches:
            return

        label_cls_add = (
            "--info-error" if msg_type == "error" else "--info-success"
        )
        label_cls_remove = (
            "--info-success" if msg_type == "error" else "--info-error"
        )

        label.update(value)
        label.set_class(True, label_cls_add)
        label.set_class(False, label_cls_remove)
        label.visible = True

        self.has_message = True

    def on_select_changed(self, event: Select.Changed) -> None:
        self.clear_message()

        select = event.select

        if not select.id == "auth-type" or select.value == self.auth_type:
            return

        self.auth_type = select.value

    def watch_auth_type(self, new: Literal["token", "userpass"]) -> None:
        labels = {"token": "API Token:", "userpass": "User / Pass:"}
        placeholders = {
            "token": "USER@REALM!TOKENID=UUID",
            "userpass": "username:password",
        }

        try:
            label = self.query_one("#auth-label", Label)
            input = self.query_one("#auth-input", Input)
        except NoMatches:
            return

        label.update(labels[new])
        input.placeholder = placeholders[new]

    def watch_hypervisor_valid(self, old: bool, new: bool) -> None:
        if old == new:
            return

        try:
            save = self.query_one("#save", Button)
            validate = self.query_one("#validate", Button)
        except NoMatches:
            return

        save.disabled = not new
        validate.disabled = new



from __future__ import annotations

import asyncio
import json
import urllib.parse
from contextlib import asynccontextmanager
from dataclasses import dataclass
from io import BufferedReader
from pathlib import Path
from socket import AF_INET
from time import monotonic
from typing import Any, AsyncIterator, Literal

import aiohttp
from aiohttp import (
    ClientError,
    ClientResponseError,
    ClientTimeout,
    ConnectionTimeoutError,
    ContentTypeError,
    TCPConnector,
)


@dataclass
class ApiResponse:
    status: int | None = None
    payload: Any = None
    timed_out: bool = False
    error: str | None = None

    @property
    def unauthorized(self) -> bool:
        return self.status == 401

    def __bool__(self) -> bool:
        return self.status == 200 and not self.error


class ProxmoxApi:
    @classmethod
    async def from_user_pass(
        cls, url: str, user: str, password: str
    ) -> ProxmoxApi | None:
        ticket_url = f"{url}/api2/json/access/ticket"
        payload = {"username": f"{user}@pam", "password": password}

        conn = TCPConnector(family=AF_INET)
        timeout = ClientTimeout(connect=2)

        try:
            async with aiohttp.ClientSession(
                connector=conn, timeout=timeout
            ) as client:
                res = await client.post(ticket_url, data=payload, ssl=False)
        except ClientError:
            return None

        try:
            res.raise_for_status()
        except ClientResponseError:
            return None

        try:
            data: dict = await res.json()
        except (json.JSONDecodeError, ContentTypeError):
            return None

        try:
            ticket = data["data"]["ticket"]
        except KeyError:
            return None

        if not ticket:
            return None

        try:
            csrf = data["data"]["CSRFPreventionToken"]
        except KeyError:
            return None

        if not csrf:
            return None

        client = cls.build_password_client(url, ticket, csrf)

        return cls(auth_type="userpass", client=client)

    @classmethod
    def parse_user_pass(cls, user_pass: str) -> tuple[str, str] | None:
        try:
            user, password = user_pass.split(":")
        except ValueError:
            return None

        # normalize, we add it back on later
        user = user.removesuffix("@pam")

        return user, password

    @classmethod
    def parse_token(cls, token: str) -> tuple[str, str, str] | None:
        try:
            user, token_parts = token.split("!")
        except ValueError:
            return None

        try:
            token_name, token_value = token_parts.split("=")
        except ValueError:
            return None

        return user, token_name, token_value

    @classmethod
    def from_token(
        cls,
        url: str,
        user: str,
        token_name: str,
        token_value: str,
    ) -> ProxmoxApi:
        client = cls.build_token_client(url, user, token_name, token_value)

        return cls(auth_type="token", client=client)

    @staticmethod
    def build_password_client(
        url: str, ticket: str, csrf_token: str
    ) -> aiohttp.ClientSession:
        cookies = {"PVEAuthCookie": ticket}
        headers = {"CSRFPreventionToken": csrf_token}

        jar = aiohttp.CookieJar(quote_cookie=False)
        jar.update_cookies(cookies)

        return aiohttp.ClientSession(
            base_url=f"{url}/api2/json/",
            cookie_jar=jar,
            cookies=cookies,
            headers=headers,
        )

    @staticmethod
    def build_token_client(
        url: str, user: str, token_name: str, token_value: str
    ) -> aiohttp.ClientSession:
        headers = {
            "Authorization": f"PVEAPIToken={user}!{token_name}={token_value}"
        }
        conn = TCPConnector(family=AF_INET)
        timeout = ClientTimeout(connect=2)

        return aiohttp.ClientSession(
            base_url=f"{url}/api2/json/",
            connector=conn,
            headers=headers,
            timeout=timeout,
        )

    @staticmethod
    async def handle_api_response(
        response: aiohttp.ClientResponse,
    ) -> ApiResponse:
        try:
            response.raise_for_status()
        except ClientResponseError as e:
            return ApiResponse(status=e.status, error=e.message)

        try:
            payload: dict = await response.json()
        except (json.JSONDecodeError, ContentTypeError) as e:
            return ApiResponse(status=response.status, error=str(e))

        data = payload.get("data", None)
        message = payload.get("message", None)

        return ApiResponse(status=response.status, payload=data, error=message)

    def __init__(
        self,
        auth_type: Literal["token", "userpass"],
        client: aiohttp.ClientSession,
    ) -> None:
        self.type = auth_type
        self.client = client

    @asynccontextmanager
    async def session(self) -> AsyncIterator[None]:
        try:
            yield
        finally:
            await self.close()

    async def close(self) -> None:
        await self.client.close()

    async def do_http(
        self,
        verb: Literal["get", "put", "post", "delete"],
        path: str,
        data: Any = None,
    ) -> ApiResponse:
        method = getattr(self.client, verb)

        try:
            res = await method(path, ssl=False, data=data)
        except ConnectionTimeoutError:
            return ApiResponse(timed_out=True)
        except ClientError:
            return ApiResponse(error="Connection Error")

        data = await self.handle_api_response(res)

        print(res.url, data)

        return data

    async def do_get(self, path: str) -> ApiResponse:
        data = await self.do_http("get", path)

        return data

    async def do_post(self, path: str, data: Any = None) -> ApiResponse:
        data = await self.do_http("post", path, data)

        return data

    async def do_put(self, path: str, data: Any = None) -> ApiResponse:
        data = await self.do_http("put", path, data)

        return data

    async def do_delete(self, path: str) -> ApiResponse:
        data = await self.do_http("delete", path)

        return data

    async def get_api_version(self, node: str) -> ApiResponse:
        res = await self.do_get(f"nodes/{node}/version")

        return res

    async def get_cluster_nodes(self) -> ApiResponse:
        res = await self.do_get("cluster/config/nodes")

        return res

    async def get_hypervisor_nodes(self) -> ApiResponse:
        res = await self.do_get("nodes")

        return res

    async def get_networks(self, node: str) -> ApiResponse:
        endpoint = f"nodes/{node}/network"

        res = await self.do_get(endpoint)

        return res

    async def get_storage_state(self, node: str) -> ApiResponse:
        endpoint = f"nodes/{node}/storage"

        res = await self.do_get(endpoint)

        return res

    async def get_storage_content(
        self, node: str, location: str
    ) -> ApiResponse:
        endpoint = f"nodes/{node}/storage/{location}/content"

        res = await self.do_get(endpoint)

        return res

    async def get_storage_config(self) -> ApiResponse:
        res = await self.do_get("storage")

        return res

    async def set_storage_content(
        self, storage_id: str, content: str
    ) -> ApiResponse:
        endpoint = f"storage/{storage_id}"
        data = {"content": content}

        res = await self.do_put(endpoint, data=data)

        return res

    async def download_iso(
        self, url: str, filename: str, storage: str, verify_certs: bool = True
    ) -> ApiResponse:
        data = {
            "content": "iso",
            "filename": filename,
            "storage": storage,
            "url": url,
            "verify-certificates": verify_certs,
        }

        res = await self.do_post(url, data=data)

        return res

    async def get_next_id(self) -> ApiResponse:
        res = await self.do_get("cluster/nextid")

        return res

    async def create_vm(self, config: dict, node: str) -> ApiResponse:
        res = await self.do_post(f"nodes/{node}/qemu", data=config)

        return res

    async def start_vm(self, vm_id: int, node: str) -> ApiResponse:
        endpoint = f"nodes/{node}/qemu/{vm_id}/status/start"

        res = await self.do_post(endpoint)

        return res

    async def get_task(self, task_id: str, node: str) -> ApiResponse:
        quoted_task = urllib.parse.quote(task_id)
        endpoint = f"nodes/{node}/tasks/{quoted_task}/status"

        res = await self.do_get(endpoint)

        return res

    async def wait_for_task(
        self, task_id: str, node: str, max_wait_s: int = 10
    ) -> bool:
        task_res = await self.get_task(task_id, node)

        if not task_res:
            return False

        exit_status = task_res.payload.get("exitstatus")
        # we start the timer here, so we don't include the time it took
        # to get the first api request
        start = monotonic()
        elapsed = 0.0

        while exit_status != "OK" and elapsed < max_wait_s:
            await asyncio.sleep(1)

            task_res = await self.get_task(task_id, node)

            if not task_res:
                return False

            exit_status = task_res.payload.get("exitstatus")
            elapsed = monotonic() - start

        return exit_status == "OK"

    async def delete_file(
        self, file_name: str, node: str, storage: str, content: str
    ) -> ApiResponse:
        volume = f"{storage}:{content}/{file_name}"
        endpoint = f"nodes/{node}/storage/{storage}/content/{volume}"

        res = await self.do_delete(endpoint)

        return res

    async def upload_file(
        self,
        file: Path | bytes,
        node: str,
        storage: str,
        file_name: str | None = None,
    ) -> ApiResponse:
        endpoint = f"nodes/{node}/storage/{storage}/upload"
        payload: BufferedReader | bytes

        if isinstance(file, Path):
            payload = open(file, "rb")
        elif (is_bytes := isinstance(file, bytes)) and not file_name:
            return ApiResponse(error="Unuseable file")
        elif is_bytes:
            payload = file
        else:
            return ApiResponse(error="Unuseable file")

        # Why proxmox does this is beyond me. This took hours to figure out. Had to
        # packet capture and decode the ssl traffic with wireshark. The multipart data
        # has to be in this specific order, as well as the headers have to be in order.
        with aiohttp.MultipartWriter("form-data") as mpwriter:
            content_part = mpwriter.append(
                "import".encode("utf-8"),
            )
            content_part.set_content_disposition("form-data", name="content")

            file_part = mpwriter.append(
                payload,
            )

            # The name=filename must be present
            file_part.set_content_disposition(
                "form-data", name="filename", filename=file_name or file.name
            )

            for part in mpwriter:
                # this is because proxmox is insane, the content-disposition header
                # must be before the content-type. Absolutely cooked.
                content_type = part[0].headers.pop(aiohttp.hdrs.CONTENT_TYPE)
                part[0].headers[aiohttp.hdrs.CONTENT_TYPE] = content_type

            res = await self.do_post(
                endpoint,
                data=mpwriter,
            )

        return res

    async def get_vms(self, node: str) -> ApiResponse:
        endpoint = f"nodes/{node}/qemu"

        res = await self.do_get(endpoint)

        return res

    async def get_vm(self, vm_id: int, node: str) -> ApiResponse:
        endpoint = f"nodes/{node}/qemu/{vm_id}/config"

        res = await self.do_get(endpoint)

        return res



from __future__ import annotations

from functools import partial
from pathlib import Path

from textual import on, work
from textual.app import App
from textual.worker import Worker, WorkerCancelled

from .messages import ScreenRequested, UpdateDefaultPage
from .models import (
    ArcaneCreatorConfig,
    ArcaneOsConfig,
    ArcaneOsConfigGroup,
    HypervisorConfig,
)
from .screens import (
    AddHypervisorScreen,
    ConfirmScreen,
    ProvisioningInfoScreen,
    ProvisionNodeScreen,
    WelcomeScreen,
    WelcomeScreenProxmox,
)


class ArcaneMage(App):
    CSS_PATH = "arcane_mage.tcss"
    AUTO_FOCUS = None
    SCREENS = {
        "welcome": WelcomeScreen,
    }

    def __init__(self, fluxnode_config: str) -> None:
        super().__init__()

        self.config = ArcaneCreatorConfig.from_fs()
        self.fluxnodes = ArcaneOsConfigGroup.from_fs(Path(fluxnode_config))

    def on_mount(self) -> None:
        self.push_screen("welcome")

        if self.config.default_page == "welcome-proxmox":
            screen = WelcomeScreenProxmox(
                self.config.hypervisors, self.fluxnodes, is_homepage=True
            )
            self.install_screen(screen, name="welcome-proxmox")
            self.push_screen("welcome-proxmox")

    def hypervisor_callback(self, hypervisor: HypervisorConfig | None) -> None:
        if not hypervisor:
            return

        # the welcome-proxmox screen has the same hypervisor list
        added = self.config.add_hypervisor(hypervisor)

        if not added:
            self.notify("Unable to add Hypervisor (keyring)")
            return

        screen = self.get_screen("welcome-proxmox", WelcomeScreenProxmox)
        screen.validate_hypervisors(hypervisor)

    def del_hypervisor_callback(
        self, hypervisor: HypervisorConfig, delete: bool
    ) -> None:
        if not delete:
            return

        self.config.remove_hypervisor(hypervisor)
        screen = self.get_screen("welcome-proxmox", WelcomeScreenProxmox)
        screen.validate_hypervisors()

    async def provision_node_callback(
        self, fluxnode: ArcaneOsConfig | None, delete_efi: bool = True
    ) -> Worker | None:
        if not fluxnode or not fluxnode.hypervisor:
            return None

        vm_name = fluxnode.hypervisor.vm_name
        total_steps = 9 if fluxnode.hypervisor.start_on_creation else 8
        info_screen = ProvisioningInfoScreen(
            vm_name=vm_name, total_steps=total_steps
        )

        # this is a bit rugged
        mounted = self.push_screen(info_screen)
        await mounted

        screen = self.get_screen("welcome-proxmox", WelcomeScreenProxmox)
        worker = screen.provision_node(
            fluxnode,
            callback=info_screen.update_callback,
            delete_efi=delete_efi,
        )

        return worker

    @on(ScreenRequested)
    def on_screen_requested(self, event: ScreenRequested) -> None:
        if event.screen == "welcome-proxmox" and not self.is_screen_installed(
            "welcome-proxmox"
        ):
            screen = WelcomeScreenProxmox(
                self.config.hypervisors, self.fluxnodes
            )
            self.install_screen(screen, name="welcome-proxmox")
        elif event.screen != "welcome-proxmox":
            self.notify("Not implemented yet")
            return

        self.push_screen(event.screen)

    @on(UpdateDefaultPage)
    def on_update_default_page(self, event: UpdateDefaultPage) -> None:
        default = event.page if event.value else None

        # this is blocking but meh
        self.config.update_default_page(default)

    @on(WelcomeScreenProxmox.AddHypervisor)
    def on_add_hypervisor(self) -> None:
        self.push_screen(
            AddHypervisorScreen(self.config.use_keyring),
            self.hypervisor_callback,
        )

    @on(WelcomeScreenProxmox.DelHypervisor)
    def on_del_hypervisor(
        self, event: WelcomeScreenProxmox.DelHypervisor
    ) -> None:
        dialog = f"Are you sure?\n\n This will delete hypervisor: {event.hypervisor.url}"
        self.push_screen(
            ConfirmScreen(dialog),
            callback=partial(self.del_hypervisor_callback, event.hypervisor),
        )

    @on(WelcomeScreenProxmox.ProvisionNode)
    def on_provision_node(
        self, event: WelcomeScreenProxmox.ProvisionNode
    ) -> None:
        self.push_screen(
            ProvisionNodeScreen(event.fluxnode),
            self.provision_node_callback,
        )

    @on(WelcomeScreenProxmox.ProvisionAllNodes)
    def on_provision_all_nodes(
        self, event: WelcomeScreenProxmox.ProvisionAllNodes
    ) -> None:
        self.provision_all_nodes(event.delay)

    @work(name="provision_all_nodes")
    async def provision_all_nodes(self, delay: int) -> None:
        screen = self.get_screen("welcome-proxmox", WelcomeScreenProxmox)

        provisionable_nodes = screen.get_provisionable_nodes()

        if not provisionable_nodes.first:
            return

        configured_node = await self.push_screen(
            ProvisionNodeScreen(provisionable_nodes.first, all=True),
            wait_for_dismiss=True,
        )

        if not configured_node:
            return

        hashed_password = configured_node.system.hashed_console

        worker = await self.provision_node_callback(configured_node)

        if not worker:
            return

        result = await worker.wait()

        if not result:
            return

        needs_pop = len(provisionable_nodes) > 1

        if delay and not configured_node == provisionable_nodes.last:
            worker = self.screen.show_delay(delay)

            try:
                await worker.wait()
            except WorkerCancelled:
                # if the user clicks okay... the screen will be dismissed
                needs_pop = False

        if needs_pop:
            # this is the info screen for the first node
            self.pop_screen()

        # we leave the info screen to display on error, otherwise pop
        for fluxnode in provisionable_nodes.rest:
            needs_pop = True
            fluxnode.system.hashed_console = hashed_password

            last = fluxnode == provisionable_nodes.last

            worker = await self.provision_node_callback(
                fluxnode, delete_efi=last
            )

            if not worker:
                break

            result = await worker.wait()

            if not result:
                break

            if delay and not last:
                worker = self.screen.show_delay(delay)

                try:
                    await worker.wait()
                except WorkerCancelled:
                    needs_pop = True

            if needs_pop:
                self.pop_screen()



from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.screen import ModalScreen
from textual.widgets import Button, Label


class ConfirmScreen(ModalScreen[bool]):
    def __init__(self, action: str) -> None:
        super().__init__()

        self.action = action

    def compose(self) -> ComposeResult:
        with Container():
            yield Label(
                self.action,
                id="info-label",
            )

            with Horizontal(id="button-container"):
                yield Button("Cancel", id="cancel")
                yield Button("Ok", id="ok")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        confirmed = event.button.id == "ok"
        self.dismiss(confirmed)


if __name__ == "__main__":
    from textual.app import App
    from textual.screen import Screen

    class Dummy(Screen):
        def compose(self) -> ComposeResult:
            yield Label("background")

    class TestApp(App):
        CSS_PATH = "../frontend.tcss"

        def on_mount(self) -> None:
            self.push_screen(Dummy())
            self.push_screen(ConfirmScreen("Are you sure?"))

    app = TestApp()
    app.run()



from rich.spinner import Spinner as RichSpinner
from textual.app import RenderableType
from textual.widgets import Static


class IntervalUpdater(Static):
    def __init__(self, renderable_object: RenderableType) -> None:
        super().__init__(renderable_object)

    def on_mount(self) -> None:
        self.interval_update = self.set_interval(1 / 15, self.refresh)

    def stop(self) -> None:
        self.interval_update.stop()


class Spinner(IntervalUpdater):
    def __init__(self, style: str) -> None:
        super().__init__(RichSpinner(style))



from __future__ import annotations

from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.screen import Screen
from textual.widgets import Button, Footer, Header, Label

from arcane_mage.messages import ScreenRequested


class WelcomeScreen(Screen):
    TITLE = "Arcane Creator"

    def __init__(
        self,
    ) -> None:
        super().__init__()

    def on_screen_resume(self):
        self.app.set_focus(None)

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)
        with Container():
            with Vertical():
                yield Label(
                    "Welcome to the Arcane Creator. Select your operation mode."
                )
            with Horizontal():
                yield Button(
                    "Proxmox Provisioner",
                    id="welcome-proxmox",
                    tooltip="Provision a Proxmox Arcane Fluxnode",
                    classes="action-button",
                )
                yield Button(
                    "Multicast Provisioner",
                    id="welcome-multicast",
                    tooltip="Configure an Arcane Fluxnode via Multicast",
                    classes="action-button",
                )
                yield Button(
                    "Usb Stick Creator",
                    id="welcome-usb",
                    tooltip="Provision a Bare Metal Arcane Fluxnode",
                    classes="action-button",
                )
            with Horizontal():
                yield Button("Exit", id="exit")
        yield Footer()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if not event.button.id:
            return

        if event.button.id == "exit":
            self.app.exit()

        self.post_message(ScreenRequested(event.button.id))



from __future__ import annotations

import asyncio
import gzip
import importlib.resources as resources
import tempfile
import uuid
from pathlib import Path
from typing import Any, Callable, Literal

import warnings

# fix this before it becomes an error (11/25)
# https://github.com/PyFilesystem/pyfilesystem2/pull/590
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=UserWarning)
    import fs

import yaml
from textual import work
from textual.app import ComposeResult
from textual.containers import Container, Grid, Horizontal, Vertical
from textual.css.query import NoMatches
from textual.message import Message
from textual.reactive import var
from textual.screen import Screen
from textual.widgets import (
    Button,
    DataTable,
    Footer,
    Header,
    Input,
    Label,
    Rule,
    Select,
    Switch,
)
from yarl import URL

from ..helpers import do_http
from ..messages import (
    UpdateDefaultPage,
)
from ..models import ArcaneOsConfig, ArcaneOsConfigGroup, HypervisorConfig
from ..proxmox import ProxmoxApi
from ..log import log


class WelcomeScreenProxmox(Screen):
    class AddHypervisor(Message): ...

    class DelHypervisor(Message):
        def __init__(self, hypervisor: HypervisorConfig):
            super().__init__()
            self.hypervisor = hypervisor

    class ProvisionNode(Message):
        def __init__(self, fluxnode: ArcaneOsConfig) -> None:
            super().__init__()

            self.fluxnode = fluxnode

    class ProvisionAllNodes(Message):
        def __init__(self, delay: int) -> None:
            super().__init__()

            self.delay = delay

    TITLE = "Proxmox Arcane Fluxnode Creator"

    first_time = var(False)
    hypervisor_populated = var(False)
    display_table = var(False)

    _images_ref = resources.files("arcane_mage.images")

    # References to compressed images in package
    efi_gz_resource = _images_ref / "arcane_efi.raw.gz"
    config_gz_resource = _images_ref / "arcane_config.raw.gz"
    config_image_base = "arcane_config"

    # The config location on the fluxnode
    config_file_path = Path("/arcane_config.yaml")

    config_tier_map = {
        "cumulus": {"memory_mb": "8192", "scsi_gb": "220", "cpu_cores": 4},
        "nimbus": {"memory_mb": "32768", "scsi_gb": "440", "cpu_cores": 8},
        "stratus": {"memory_mb": "65536", "scsi_gb": "880", "cpu_cores": 16},
    }

    proxmox_api: ProxmoxApi

    @staticmethod
    def is_api_min_version(version: str) -> bool:
        MIN_VERSION = [8, 4, 1]

        parts = version.split(".")

        if not len(parts) == 3:
            return False

        for actual_str, required in zip(parts, MIN_VERSION):
            try:
                actual = int(actual_str)
            except ValueError:
                return False

            if actual > required:
                return True
            elif actual < required:
                return False

        return True

    def __init__(
        self,
        hypervisors: list[HypervisorConfig],
        fluxnodes: ArcaneOsConfigGroup,
        is_homepage: bool = False,
    ) -> None:
        super().__init__()

        self.hypervisors = hypervisors
        self.fluxnodes = fluxnodes
        self.is_homepage = is_homepage

        self.latest_iso: str | None = None

    def on_screen_resume(self):
        self.app.set_focus(None)

    def on_load(self) -> None:
        self.get_latest_iso_version()

    def on_mount(self) -> None:
        print(self.fluxnodes)

        self.validate_hypervisors()

    def validate_hypervisors(
        self, hypervisor: HypervisorConfig | None = None
    ) -> None:
        self.first_time = not bool(self.hypervisors)
        self.hypervisor_populated = False

        select = self.query_one(Select)

        options = [(URL(x.url).host, x.url) for x in self.hypervisors]

        if hypervisor:
            value = hypervisor.url
        else:
            value = options[0][1] if options else Select.BLANK

        select.set_options(options)
        select.value = value

    @work(name="populate_hypervisor")
    async def populate_hypervisor_api(
        self, hypervisor: HypervisorConfig
    ) -> None:
        url = hypervisor.url
        api = None

        useable_nodes = ArcaneOsConfigGroup()
        provisioned: dict[str, list] = {}

        credential = hypervisor.real_credential()

        if not credential:
            self.build_fluxnode_table(useable_nodes, provisioned)
            self.hypervisor_populated = True
            self.notify("Unable to find credential for Hypervisor")
            return

        if hypervisor.auth_type == "token":
            if token := ProxmoxApi.parse_token(credential):
                api = ProxmoxApi.from_token(url, *token)
        elif hypervisor.auth_type == "userpass":
            if user_pass := ProxmoxApi.parse_user_pass(credential):
                api = await ProxmoxApi.from_user_pass(url, *user_pass)

        if not api:
            self.build_fluxnode_table(useable_nodes, provisioned)
            self.hypervisor_populated = True
            self.notify("No Proxmox API available", severity="error")
            return

        self.proxmox_api = api

        hyper_nodes = await self.proxmox_api.get_hypervisor_nodes()

        if not hyper_nodes:
            self.build_fluxnode_table(useable_nodes, provisioned)
            self.hypervisor_populated = True
            msg = "Timeout" if hyper_nodes.timed_out else hyper_nodes.error
            self.notify(
                f"Error getting Hypervisor nodes via api: {msg}",
                severity="error",
            )
            return

        get_vms_coros = []

        async def handle_node(node: dict):
            if name := node.get("node", None):
                vm_res = await self.proxmox_api.get_vms(name)
                provisioned[name] = vm_res.payload
                useable_nodes.add_nodes(
                    self.fluxnodes.get_nodes_by_hypervisor_name(name)
                )

        for node in hyper_nodes.payload:
            get_vms_coros.append(handle_node(node))

        await asyncio.gather(*get_vms_coros)

        self.build_fluxnode_table(useable_nodes, provisioned)
        self.hypervisor_populated = True

    def build_fluxnode_table(
        self, fluxnodes: ArcaneOsConfigGroup, provisioned_nodes: dict[str, Any]
    ):
        try:
            table = self.query_one(DataTable)
        except NoMatches:
            return

        columns = [
            {"label": "Node", "key": "node"},
            {"label": "Hostname", "key": "hostname"},
            {"label": "Tier", "key": "tier"},
            {"label": "Network", "key": "network"},
            {"label": "Address", "key": "address"},
            {"label": "Provisioned", "key": "provisioned"},
        ]

        table.clear(columns=True)
        for column in columns:
            table.add_column(**column)

        # this is feral
        for fluxnode in fluxnodes:
            hyper = fluxnode.hypervisor
            row_key = f"{hyper.node}:{hyper.vm_name}"

            hypervisor_nodes = provisioned_nodes.get(hyper.node, [])

            is_provisioned = bool(
                next(
                    filter(
                        lambda x: x.get("name") == hyper.vm_name,
                        hypervisor_nodes,
                    ),
                    None,
                )
            )
            table.add_row(*fluxnode.as_row(), is_provisioned, key=row_key)

    def compose(self) -> ComposeResult:
        first_time_dialog = Label(
            'It looks like this is your first time here, click "Add" to get started.',
            id="dialog",
        )
        first_time_dialog.display = False

        fluxnode_dt: DataTable = DataTable(cursor_type="row")
        fluxnode_dt.border_title = "Fluxnodes"
        fluxnode_dt.display = False

        yield Header(show_clock=True)
        with Container():
            with Grid():
                with Horizontal(id="first-column"):
                    yield Button(
                        "\u21b0",
                        id="back",
                        classes="icon-button",
                        tooltip="Back",
                    )
                    yield Button(
                        "Add", id="add-hypervisor", tooltip="Add Hypervisor"
                    )
                    yield Button(
                        "Del", id="del-hypervisor", tooltip="Delete Hypervisor"
                    )
                with Horizontal(id="second-column"):
                    yield Label("Selected:", classes="text-label")
                    yield Select([])
                with Horizontal(id="third-column"):
                    yield Label("Homepage:", classes="text-label")
                    yield Switch(
                        id="homepage",
                        value=self.is_homepage,
                        tooltip="Set as default Homepage",
                    )
                    yield Button(
                        "X", id="exit", classes="icon-button", tooltip="Exit"
                    )
            yield Rule()
            with Vertical():
                yield first_time_dialog
                with Vertical(id="dt-container"):
                    yield fluxnode_dt
                    with Horizontal():
                        yield Label("Delay:", classes="text-label")
                        yield Input(
                            "300",
                            tooltip="Time between configuring",
                            restrict=r"^\d{0,4}",
                        )
                        yield Button("Sync All", id="sync-all")

        yield Footer()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if not event.button.id:
            return

        if event.button.id == "add-hypervisor":
            self.post_message(WelcomeScreenProxmox.AddHypervisor())
        elif event.button.id == "del-hypervisor":
            select = self.query_one(Select)
            if select.value == Select.BLANK:
                return

            if hyper := self.get_hypervisor_by_url(select.value):
                self.post_message(WelcomeScreenProxmox.DelHypervisor(hyper))

        elif event.button.id == "back":
            self.dismiss()
        elif event.button.id == "exit":
            self.app.exit()
        elif event.button.id == "sync-all":
            input = self.query_one(Input)
            delay = int(input.value) if input.value else 0
            self.post_message(WelcomeScreenProxmox.ProvisionAllNodes(delay))

    def on_switch_changed(self, event: Switch.Changed) -> None:
        if event.switch.id != "homepage":
            return

        self.post_message(
            UpdateDefaultPage("welcome-proxmox", event.switch.value)
        )

    def get_provisionable_nodes(self) -> ArcaneOsConfigGroup:
        # stop doing this. Store the data, so we don't have to get it from the dt
        try:
            dt = self.query_one(DataTable)
        except NoMatches:
            return ArcaneOsConfigGroup()

        nodes: list[ArcaneOsConfig] = []

        for row_key in dt.rows:
            row = dt.get_row(row_key)

            if not row[-1]:  # not provisioned
                fluxnode = self.fluxnodes.get_node_by_vm_name(
                    *row_key.value.split(":")
                )

                if fluxnode:
                    nodes.append(fluxnode)

        return ArcaneOsConfigGroup(nodes)

    def on_data_table_row_selected(self, event: DataTable.RowSelected) -> None:
        row = event.data_table.get_row(event.row_key)

        provisioned = row.pop()

        if provisioned:
            name = row[1]
            self.notify(f"Fluxnode: {name} already provisioned")
            return

        # fix this
        fluxnode = self.fluxnodes.get_node_by_vm_name(
            *event.row_key.value.split(":")
        )

        if not fluxnode:
            return

        self.post_message(WelcomeScreenProxmox.ProvisionNode(fluxnode))

    async def validate_api_version(self, node: str) -> tuple[bool, str]:
        res = await self.proxmox_api.get_api_version(node)

        if not res:
            return False, "Unable to get Proxmox api version"

        version = res.payload.get("version")

        if not version:
            return False, "Api payload missing version info"

        if not self.is_api_min_version(version):
            return False, f"Api version too old. Got: {version}, Want: 8.4.1"

        return True, ""

    async def validate_storage(
        self,
        node: str,
        storage_iso: str,
        storage_images: str,
        storage_import: str,
    ) -> tuple[bool, str]:
        res = await self.proxmox_api.get_storage_state(node)

        if not res:
            return False, "Unable to get Proxmox storage state"

        if not res.payload:
            return (
                False,
                "No Storage state available, did you forget API permissions?",
            )

        node_storage_iso = next(
            filter(lambda x: x.get("storage") == storage_iso, res.payload), None
        )

        node_storage_images = next(
            filter(lambda x: x.get("storage") == storage_images, res.payload),
            None,
        )

        node_storage_import = next(
            filter(lambda x: x.get("storage") == storage_import, res.payload),
            None,
        )

        if not all(
            [node_storage_iso, node_storage_images, node_storage_import]
        ):
            return False, "Missing storage config item"

        iso_content = node_storage_iso.get("content")
        images_content = node_storage_images.get("content")
        import_content = node_storage_import.get("content")

        if (
            "iso" not in iso_content
            or "images" not in images_content
            or "import" not in import_content
        ):
            return False, "Storage type missing on hypervisor"

        import_available = node_storage_import.get("avail", 0)

        # We need 4MiB + 4MiB for the EFI image and the config image. So we check for 10MiB
        if import_available < 10485760:
            return False, "Storage space less than 10MiB on hypervisor"

        # Should really parse these into storage objects and do stuff with them

        # {
        #     "data": [
        #         {
        #         "used": 1561132007424,
        #         "total": 3651770777600,
        #         "content": "images,rootdir",
        #         "enabled": 1,
        #         "avail": 2090638770176,
        #         "storage": "local-lvm",
        #         "used_fraction": 0.4275,
        #         "type": "lvmthin",
        #         "active": 1,
        #         "shared": 0
        #         },
        #         {
        #         "type": "dir",
        #         "active": 1,
        #         "shared": 0,
        #         "used_fraction": 0.941026375512876,
        #         "storage": "local",
        #         "avail": 2715500544,
        #         "enabled": 1,
        #         "total": 206551633920,
        #         "content": "import,backup,iso,vztmpl",
        #         "used": 194370535424
        #         }
        #     ]
        # }

        return True, ""

    async def validate_iso_version(
        self, node: str, iso_name: str, storage_iso: str
    ) -> bool:
        res = await self.proxmox_api.get_storage_content(node, storage_iso)

        if not res:
            return False

        iso_exists = next(
            filter(
                lambda x: x.get("content") == "iso"
                and x.get("volid", "").endswith(iso_name),
                res.payload,
            ),
            None,
        )

        return bool(iso_exists)

    async def validate_network(self, node: str, network: str) -> bool:
        res = await self.proxmox_api.get_networks(node)

        if not res:
            return False

        network_exists = next(
            filter(lambda x: x.get("iface") == network, res.payload), None
        )

        return bool(network_exists)

    def get_hypervisor_by_url(self, url: str) -> HypervisorConfig | None:
        return next(filter(lambda x: x.url == url, self.hypervisors), None)

    def on_select_changed(self, event: Select.Changed) -> None:
        if (hyper_url := event.select.value) == Select.BLANK:
            return

        if hypervisor := self.get_hypervisor_by_url(hyper_url):
            self.populate_hypervisor_api(hypervisor)

    @work(name="get_latest_iso_version")
    async def get_latest_iso_version(self) -> None:
        url = "https://images.runonflux.io/api/latest_release"

        res = await do_http(url, total_timeout=3)

        if not res:
            return None

        self.latest_iso = res.get("iso", None)

    @work(name="provision_node")
    async def provision_node(
        self,
        fluxnode: ArcaneOsConfig,
        callback: Callable[[bool, str], None],
        delete_efi: bool = True,
    ) -> bool:
        hv = fluxnode.hypervisor

        if not hv:
            return False

        if hv.node_tier not in WelcomeScreenProxmox.config_tier_map:
            callback(False, f"Node tier: {hv.node_tier} does not exist")
            return False

        version_valid, version_error = await self.validate_api_version(hv.node)

        if not version_valid:
            callback(False, version_error)
            return False

        callback(True, "Api version validated")

        storage_valid, storage_error = await self.validate_storage(
            hv.node, hv.storage_iso, hv.storage_images, hv.storage_import
        )

        if not storage_valid:
            callback(False, storage_error)
            return False

        callback(True, "Storage validated")

        iso_valid = await self.validate_iso_version(
            hv.node, hv.iso_name, hv.storage_iso
        )

        if not iso_valid:
            callback(False, "Unable to find ISO image on hypervisor")
            return False

        callback(True, "ISO image validated")

        network_valid = await self.validate_network(hv.node, hv.network)

        if not network_valid:
            callback(False, "Network not present on hypervisor")
            return False

        callback(True, "Network validated")

        vm_config = await self.create_vm_config(
            vm_name=hv.vm_name,
            vm_id=hv.vm_id,
            tier=hv.node_tier,
            network_bridge=hv.network,
            iso_name=hv.iso_name,
            disk_limit=hv.disk_limit,
            cpu_limit=hv.cpu_limit,
            startup_config=hv.startup_config,
        )

        if not vm_config:
            callback(False, "Unable to generate vm config")
            return False

        vm_id: int = vm_config["vmid"]

        config_upload = yaml.dump({"nodes": [fluxnode.to_dict()]})

        config_ok = await self.upload_arcane_config(
            config_upload.encode("utf-8"), vm_id, hv.node, hv.storage_import
        )

        print("CONFIG OK:", config_ok)

        if not config_ok:
            callback(False, "Unable to upload Config image to hypervisor")
            return False

        callback(True, "Config image uploaded")

        efi_ok = await self.upload_arcane_efi(hv.node, hv.storage_import)

        print("EFI OK:", efi_ok)

        if not efi_ok:
            callback(False, "Unable to upload EFI image to hypervisor")
            return False

        callback(True, "EFI image uploaded")

        created_ok = await self.create_vm(vm_config, node=hv.node)

        print("CREATED OK", created_ok)

        if not created_ok:
            await self.delete_install_disks(
                vm_id, hv.node, hv.storage_import, delete_efi
            )
            callback(False, "Unable to create VM on hypervisor")
            return False

        callback(True, "VM Created")

        deleted_ok = await self.delete_install_disks(
            vm_id, hv.node, hv.storage_import, delete_efi
        )

        print("DELETED OK", created_ok)

        if not deleted_ok:
            callback(False, "Unable to clean up disk images on hypervisor")
            return False

        callback(True, "Disk images cleaned")

        row_key = f"{hv.node}:{hv.vm_name}"
        table = self.query_one(DataTable)
        table.update_cell(row_key, "provisioned", True)

        if not hv.start_on_creation:
            return True

        started_ok = await self.start_vm(vm_id, hv.node)

        print("STARTED OK", created_ok)

        if not started_ok:
            callback(False, "Unable to start VM on hypervisor")
            return False

        callback(True, "VM started")
        return True

    def get_vm_config_file_name(self, vm_id: int) -> str:
        return f"{vm_id}_{self.config_image_base}.raw"

    async def start_vm(self, vm_id: int, node: str) -> bool:
        res = await self.proxmox_api.start_vm(vm_id, node)

        if not res:
            return False

        ok = await self.proxmox_api.wait_for_task(res.payload, node, 20)

        return ok

    async def create_vm(self, config: dict, node: str) -> bool:
        create_res = await self.proxmox_api.create_vm(config, node)

        if not create_res:
            log.error(create_res)
            return False

        ok = await self.proxmox_api.wait_for_task(create_res.payload, node)

        return ok

    async def delete_install_disks(
        self, vm_id: int, node: str, storage: str, delete_efi: bool = True
    ) -> bool:
        efi_file = "arcane_efi.raw"
        config_file = f"{vm_id}_arcane_config.raw"

        if delete_efi:
            efi_res = await self.proxmox_api.delete_file(
                efi_file, node, storage, content="import"
            )
        else:
            efi_res = True

        config_res = await self.proxmox_api.delete_file(
            config_file, node, storage, content="import"
        )

        if not efi_res or not config_res:
            return False

        if delete_efi:
            efi_ok = await self.proxmox_api.wait_for_task(efi_res.payload, node)
        else:
            efi_ok = True

        if not efi_ok:
            return False

        config_ok = await self.proxmox_api.wait_for_task(
            config_res.payload, node
        )

        return config_ok

    async def upload_arcane_efi(self, node: str, storage: str) -> bool:
        with self.efi_gz_resource.open("rb") as f:
            efi_disk = gzip.decompress(f.read())

        upload_res = await self.proxmox_api.upload_file(
            efi_disk,
            node=node,
            storage=storage,
            file_name="arcane_efi.raw",
        )

        if not upload_res:
            return False

        ok = await self.proxmox_api.wait_for_task(upload_res.payload, node)

        return ok

    async def upload_arcane_config(
        self, config: bytes, vm_id: int, node: str, storage: str
    ) -> bool:
        with tempfile.TemporaryDirectory(prefix="arcane_mage_") as tmpdir:
            config_image_name = self.get_vm_config_file_name(vm_id)
            config_image_path = Path(tmpdir) / config_image_name

            # Extract and decompress the config image template
            with config_image_path.open("wb") as img_fh:
                with self.config_gz_resource.open("rb") as img_gz_fh:
                    img_fh.write(gzip.decompress(img_gz_fh.read()))

            # Modify the FAT filesystem to add the config
            fat_fs = fs.open_fs(f"fat://{config_image_path}")
            with fat_fs.openbin(
                str(self.config_file_path), mode="wb"
            ) as conf_fh:
                conf_fh.write(config)

            # Upload the modified image
            upload_res = await self.proxmox_api.upload_file(
                config_image_path,
                node=node,
                storage=storage,
            )

        if not upload_res:
            return False

        ok = await self.proxmox_api.wait_for_task(upload_res.payload, node)

        return ok

    async def create_vm_config(
        self,
        vm_name: str,
        tier: Literal["cumulus", "nimbus", "stratus"],
        network_bridge: str,
        storage_images: str = "local-lvm",
        storage_iso: str = "local",
        storage_import: str = "local",
        vm_id: int | None = None,
        iso_name: str | None = None,
        startup_config: str | None = None,
        disk_limit: int | None = None,
        cpu_limit: float | None = None,
    ) -> dict | None:
        tier_config = WelcomeScreenProxmox.config_tier_map.get(tier)

        if not tier_config:
            return None

        if vm_id is None:
            vm_id_res = await self.proxmox_api.get_next_id()

            if not vm_id_res:
                return None

            vm_id = vm_id_res.payload

            assert vm_id

        disk_rate = (
            f"mbps_rd={disk_limit},mbps_wr={disk_limit}," if disk_limit else ""
        )
        cpu_limit = cpu_limit or 0

        smbios_uuid = str(uuid.uuid4())

        config_img = self.get_vm_config_file_name(vm_id)

        config = {
            "efidisk0": f"{storage_images}:0,efitype=4m,pre-enrolled-keys=0,import-from={storage_import}:import/arcane_efi.raw",
            "cpu": "host",
            "ostype": "l26",
            "sockets": 1,
            "vmid": vm_id,
            "agent": "1",
            "onboot": 1,
            "name": vm_name,
            "smbios1": f"uuid={smbios_uuid}",
            "boot": "order=scsi0;ide2;net0",
            "numa": 0,
            "memory": tier_config["memory_mb"],
            "tpmstate0": f"{storage_images}:4,version=v2.0",
            "cores": tier_config["cpu_cores"],
            "cpulimit": cpu_limit,
            "bios": "ovmf",
            "scsi0": f"{storage_images}:{tier_config['scsi_gb']},{disk_rate}discard=on,iothread=1,ssd=1",
            "scsi1": f"{storage_images}:0,import-from={storage_import}:import/{config_img}",
            "ide2": f"{storage_iso}:iso/{iso_name},media=cdrom",
            "net0": f"model=virtio,bridge={network_bridge}",
            "scsihw": "virtio-scsi-single",
        }

        if startup_config:
            config["startup"] = startup_config

        return config

    def compute_display_table(self) -> bool:
        return not self.first_time and self.hypervisor_populated

    def watch_display_table(self, old: bool, new: bool) -> None:
        if old == new:
            return

        try:
            table = self.query_one(DataTable)
        except NoMatches:
            return

        table.display = new

    def watch_first_time(self, old: bool, new: bool) -> None:
        if old == new:
            return

        try:
            label = self.query_one("#dialog", Label)
        except NoMatches:
            return

        label.display = new



from __future__ import annotations

import asyncio

from textual import work
from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.dom import NoMatches
from textual.screen import ModalScreen
from textual.widgets import Button, Label, ProgressBar
from textual.worker import Worker

from ..widgets import Spinner


class ProvisioningInfoScreen(ModalScreen):
    @property
    def finished(self) -> bool:
        return self.completed_steps >= self.total_steps

    @property
    def last_step(self) -> bool:
        return self.completed_steps + 1 == self.total_steps

    def __init__(self, vm_name: str, total_steps: int) -> None:
        super().__init__()

        self.vm_name = vm_name
        self.completed_steps = 0
        self.total_steps = total_steps

        self.failed = False

    def compose(self) -> ComposeResult:
        ok_button = Button("Ok")
        ok_button.visible = False
        error_label = Label("", id="error-label")
        error_label.visible = False
        main_container = Container()
        main_container.border_title = f"Provisioining: {self.vm_name}"

        wait_container = Horizontal(id="wait-container")
        wait_container.visible = False

        with main_container:
            with Container():
                with Vertical():
                    yield Label("", id="task-label")
                    with Horizontal():
                        yield Spinner("line")
                        yield ProgressBar(
                            show_eta=False,
                            id="task-progress",
                            total=self.total_steps,
                        )
            yield error_label
            with Horizontal(id="button-container"):
                yield ok_button
            with wait_container:
                yield Label("Countdown: 0", id="wait-label")

    def show_delay(self, delay: int) -> Worker:
        return self.wait_delay(delay)

    @work(name="wait_delay")
    async def wait_delay(self, delay: int) -> None:
        try:
            container = self.query_one("#wait-container", Horizontal)
        except NoMatches:
            return

        label = self.query_one("#wait-label", Label)
        label.update(f"Countdown: {delay}")

        container.visible = True

        while delay > 0:
            await asyncio.sleep(1)
            delay -= 1
            label.update(f"Countdown: {delay}")

    def update_callback(self, success: bool, message: str) -> None:
        if self.finished or self.failed:
            return

        try:
            progress = self.query_one(ProgressBar)
            label = self.query_one("#task-label", Label)
            spinner = self.query_one(Spinner)
            ok = self.query_one(Button)
        except NoMatches:
            return

        if success:
            progress.advance()
            label.update(message)
        else:
            self.failed = True
            label.update("Error")
            error = self.query_one("#error-label", Label)
            error.update(message)
            error.visible = True

        if not success or self.last_step:
            spinner.stop()
            spinner.visible = False
            ok.visible = True

        self.completed_steps += 1

    def on_button_pressed(self) -> None:
        self.dismiss()



from logging import getLogger
from logging.config import dictConfig

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "standard": {
            "format": "{asctime} [{levelname}] {filename} {lineno}: {message}",
            "datefmt": "%Y-%m-%d %H:%M:%S",
            "style": "{",
        },
    },
    "handlers": {
        "default": {
            "level": "INFO",
            "formatter": "standard",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",  # Default is stderr
        },
        "file_handler": {
            "level": "INFO",
            "formatter": "standard",
            "class": "logging.handlers.RotatingFileHandler",
            "filename": "arcane_mage.log",
            "mode": "a",
            "maxBytes": 1048576,
            "backupCount": 3,
        },
        "textual_handler": {
            "level": "INFO",
            "formatter": "standard",
            "class": "textual.logging.TextualHandler",
        },
        "rich_handler": {
            "level": "INFO",
            "class": "rich.logging.RichHandler",
        },
    },
    "loggers": {
        "": {  # root logger
            "handlers": [],
            "level": "WARNING",
            "propagate": False,
        },
        "arcane_mage": {
            "handlers": ["textual_handler", "file_handler"],
            "level": "INFO",
            "propagate": False,
            "__main__": {
                "handlers": ["rich_handler"],
                "level": "DEBUG",
                "propagate": False,
            },
        },
    },
}

dictConfig(LOGGING_CONFIG)

log = getLogger(__name__)



from textual.messages import Message


class ScreenRequested(Message):
    def __init__(self, screen: str) -> None:
        super().__init__()

        self.screen = screen


class UpdateDefaultPage(Message):
    def __init__(self, page: str, value: bool) -> None:
        super().__init__()

        self.page = page
        self.value = value



import secrets
from dataclasses import dataclass, field

from pyescrypt import Mode, WrongPassword, WrongPasswordConfiguration, Yescrypt


@dataclass
class HashedPassword:
    password: str
    hasher: Yescrypt = field(
        default_factory=lambda: Yescrypt(n=2**12, r=32, p=1, mode=Mode.MCF)
    )
    salt: bytes = field(default_factory=lambda: secrets.token_bytes(16))

    @property
    def encoded_pass(self) -> bytes:
        return self.password.encode()

    def hash(self) -> str:
        hashed = self.hasher.digest(password=self.encoded_pass, salt=self.salt)
        return hashed.decode()

    def validate(self, hashed: bytes) -> bool:
        try:
            self.hasher.compare(self.encoded_pass, hashed)
        except (WrongPasswordConfiguration, WrongPassword):
            return False

        return True



import click
from importlib.metadata import version as get_version

from .arcane_mage import ArcaneMage


@click.group(invoke_without_command=True)
@click.option(
    "-c", "--config", default="fluxnodes.yaml", help="The config file"
)
@click.option(
    "-v", "--version", is_flag=True, help="Show the application version"
)
@click.pass_context
def cli(ctx: click.Context, config: str, version: bool):
    if version:
        pkg_version = get_version("arcane-mage")
        click.echo(f"arcane-mage {pkg_version}")
        ctx.exit()
    if ctx.invoked_subcommand is None:
        app = ArcaneMage(fluxnode_config=config)
        app.run()


@cli.command()
def provision_proxmox():
    click.echo("Not Implemented")


@cli.command()
def provision_multicast():
    click.echo("Not Implemented")


@cli.command()
def provision_usb():
    click.echo("Not Implemented")


if __name__ == "__main__":
    cli()



from __future__ import annotations

import asyncio
import os
import pwd

from itertools import count
from json import JSONDecodeError
from socket import AF_INET
from typing import Any, Generator, Literal, AsyncGenerator
from pathlib import Path
import aiofiles
from yarl import URL

from aiohttp import (
    BasicAuth as _BasicAuth,
)
from aiohttp import (
    ClientConnectorError,
    ClientOSError,
    ClientSession,
    ClientTimeout,
    ConnectionTimeoutError,
    ContentTypeError,
    TCPConnector,
    ClientResponse,
    ServerDisconnectedError,
)

from arcane_mage.log import log


class BasicAuth(_BasicAuth):
    def __rich_repr__(self) -> Generator[tuple[str, str], None, None]:
        yield "login", self.login
        yield "password", "*****"
        yield "encoding", self.encoding


class ExecBinaryError(ChildProcessError):
    def __init__(self, cmd: list[str], stdout: bytes, stderr: bytes):
        self.__cmd = cmd
        self.__stderr = stderr
        super().__init__([cmd, (stdout, stderr)])

    def stderr(self):
        return self.__stderr.decode(errors="ignore")

    def __str__(self):
        return "Failed to execute '{}', error: {}".format(
            " ".join(self.__cmd), self.stderr()
        )


def demote(user_uid, user_gid):
    def result():
        os.setgid(user_gid)
        os.setuid(user_uid)

    return result


async def exec_binary(
    cmd: list[str],
    expect_returncode: int = 0,
    cwd: str | None = None,
    user: str | None = None,
    env: dict | None = None,
) -> str:
    if len(cmd) == 0:
        raise ChildProcessError("Cannot execute empty cmd")

    extra_params = {}

    if cwd is not None:
        extra_params["cwd"] = cwd

    if user is not None:
        try:
            pw_record = pwd.getpwnam(user)
        except KeyError:
            raise ChildProcessError("User not found")

        extra_params["preexec_fn"] = demote(pw_record.pw_uid, pw_record.pw_gid)

    if env:
        extra_params["env"] = env

    try:
        proc = await asyncio.create_subprocess_exec(
            *cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            **extra_params,
        )
    except FileNotFoundError:
        raise ChildProcessError(f"Binary not found: {cmd[0]}, is it installed?")

    try:
        stdout, stderr = await proc.communicate()
        if proc.returncode != expect_returncode:
            raise ExecBinaryError(cmd, stdout, stderr)
    except asyncio.CancelledError:
        if proc.returncode is None:
            proc.terminate()
            return_code = await proc.wait()
            log.info(
                "Cmd: %s cancelled and terminated. Exit code: %s",
                cmd[0],
                return_code,
            )
        raise asyncio.CancelledError()

    return stdout.decode()


async def do_http(
    url: URL | str,
    verb: Literal["get", "post", "head"] = "get",
    *,
    data: Any = None,
    connect_timeout: int = 3,
    max_tries: int = 3,
    retry_interval: int = 1,
    credentials: list[str] | None = None,
    headers: dict | None = None,
    verify_ssl: bool | None = None,
    total_timeout: int | None = None,
) -> Any:
    timeout = ClientTimeout(connect=connect_timeout, total=total_timeout)
    conn = TCPConnector(family=AF_INET)

    auth = None
    res = None

    if credentials:
        # user password
        auth = BasicAuth(*credentials)

    async with ClientSession(
        connector=conn,
    ) as session:
        for attempt in count(start=1):
            if max_tries and attempt > max_tries:
                break

            remaining = max(0, max_tries - attempt)

            try:
                method = getattr(session, verb)
                async with method(
                    url,
                    timeout=timeout,
                    auth=auth,
                    headers=headers,
                    json=data,
                    ssl=verify_ssl,
                ) as resp:
                    resp: ClientResponse  # type: ignore
                    if resp.status in [429, 500, 502, 503, 504]:
                        log.debug("bad response: %s", resp.status)
                        if remaining or not max_tries:
                            await asyncio.sleep(retry_interval)

                        continue

                    if resp.status in [401, 403, 404]:
                        log.info("URL: %s, status: %s", url, resp.status)
                        return None

                    if verb == "head":
                        return resp.headers

                    try:
                        res = await resp.json()
                    except (
                        JSONDecodeError,
                        ContentTypeError,
                        UnicodeDecodeError,
                    ):
                        res = None
                    break

            except (
                ClientConnectorError,
                ConnectionTimeoutError,
                ClientOSError,
                ServerDisconnectedError,
                asyncio.TimeoutError,
            ):
                # ClientOSError: have seen Connection reset by peer
                log.info(
                    "Unable to connect to: %s. Retries: %s remaining. Interval: %ss",
                    url,
                    remaining,
                    retry_interval,
                )

                if remaining or not max_tries:
                    await asyncio.sleep(retry_interval)

    return res


async def do_http_iter(
    url: str,
    verb: str = "get",
    *,
    data: Any = None,
    connect_timeout: int = 3,
    read_timeout: int = 10,
    max_tries: int = 3,
    retry_interval: int = 1,
    credentials: list | None = None,
    headers: dict | None = None,
    verify_ssl: bool | None = None,
) -> Any | AsyncGenerator:
    timeout = ClientTimeout(connect=connect_timeout, sock_read=read_timeout)
    conn = TCPConnector(family=AF_INET)

    auth = None
    ssl = None

    if credentials:
        # user password
        auth = BasicAuth(*credentials)

    if verify_ssl is not None:
        ssl = verify_ssl

    async with ClientSession(
        connector=conn,
    ) as session:
        for attempt in count(start=1):
            if max_tries and attempt > max_tries:
                break

            remaining = max(0, max_tries - attempt)

            try:
                method = getattr(session, verb)
                async with method(
                    url,
                    timeout=timeout,
                    auth=auth,
                    headers=headers,
                    json=data,
                    ssl=ssl,
                ) as resp:
                    resp: ClientResponse

                    if resp.status in [429, 500, 502, 503, 504]:
                        log.warning(
                            "URL: %s, Status: %s... retrying", url, resp.status
                        )
                        if remaining or not max_tries:
                            await asyncio.sleep(retry_interval)

                        continue

                    if resp.status in [401, 403]:
                        break

                    # This header must always be present
                    content_length = int(resp.headers.get("Content-Length"))
                    bytes_downloaded = 0
                    percent_logged = 0

                    async for chunk, _ in resp.content.iter_chunks():
                        bytes_downloaded += len(chunk)
                        percent = int((bytes_downloaded / content_length) * 100)
                        if percent > percent_logged and percent % 5 == 0:
                            percent_logged = percent
                            log.info("URL: %s, %s%% downloaded", url, percent)

                        yield chunk

                    break

            except (
                ClientConnectorError,
                ConnectionTimeoutError,
                ClientOSError,
                ServerDisconnectedError,
                asyncio.TimeoutError,
            ):
                log.warning(
                    "Unable to connect to: %s. Retries: %s remaining. Interval: %ss",
                    url,
                    remaining,
                    retry_interval,
                )

                if remaining or not max_tries:
                    await asyncio.sleep(retry_interval)


async def do_http_to_file(
    url: URL | str,
    dest_path: Path,
    connect_timeout: int = 3,
    read_timeout: int = 15,
) -> bool:
    timeout = ClientTimeout(connect=connect_timeout, sock_read=read_timeout)
    conn = TCPConnector(family=AF_INET)

    async with ClientSession(
        connector=conn,
    ) as session:
        try:
            file_handler = await aiofiles.open(dest_path, "wb")
        except PermissionError:
            return False

        try:
            async with session.get(url, timeout=timeout) as resp:
                resp: ClientResponse

                if resp.status != 200:
                    log.warning("Bad response received: %s", resp)

                    return False

                async for chunk in resp.content.iter_chunked(16777216):  # 16Mib
                    await file_handler.write(chunk)
        except (
            ClientConnectorError,
            ConnectionTimeoutError,
            ClientOSError,
            ServerDisconnectedError,
            asyncio.TimeoutError,
        ):
            # ClientOSError: have seen Connection reset by peer
            log.info(
                "Unable to connect to: %s",
                url,
            )
            return False

        finally:
            await file_handler.close()

        return True



from .confirm_screen import ConfirmScreen as ConfirmScreen
from .hypervisor import AddHypervisorScreen as AddHypervisorScreen
from .provision_node import ProvisionNodeScreen as ProvisionNodeScreen
from .provisioning_info import ProvisioningInfoScreen as ProvisioningInfoScreen
from .welcome import WelcomeScreen as WelcomeScreen
from .welcome_proxmox import WelcomeScreenProxmox as WelcomeScreenProxmox



from __future__ import annotations

from itertools import groupby

from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.reactive import var
from textual.screen import ModalScreen
from textual.validation import ValidationResult, Validator
from textual.widgets import Button, Input, Label, Markdown

from ..models import ArcaneOsConfig
from ..password import HashedPassword


class PasswordValidator(Validator):
    def __init__(self) -> None:
        super().__init__()

        self.previous: str | None = None

    def validate(self, password: str) -> ValidationResult:
        if self.previous is not None and len(password) < len(self.previous):
            self.previous = password
            return self.success()

        if len(password) < 12:
            self.previous = password
            return self.failure("Password must be a minimum of 12 characters.")

        groups = groupby(password)
        consecutives = [
            letter for letter, group in groups if len(list(group)) > 3
        ]
        if any(consecutives):
            self.previous = password
            return self.failure("Only 3 consecutive repeat characters allowed.")

        self.previous = password
        return self.success()


class ProvisionNodeScreen(ModalScreen[ArcaneOsConfig | None]):
    password_valid = var(False)

    def __init__(self, fluxnode: ArcaneOsConfig, all: bool = False) -> None:
        super().__init__()

        self.fluxnode = fluxnode
        self.all = all

    def on_screen_resume(self) -> None:
        self.query_one(Input).focus()

    def compose(self) -> ComposeResult:
        title = (
            "Provision: All nodes"
            if self.all
            else f"Provision: {self.fluxnode.hypervisor.vm_name}"
        )

        container = Container()
        container.border_title = title

        error_label = Label("", id="error-label")
        error_label.visible = False

        single_dialog = """Add console password and click Provision to configure this fluxnode.
The following will occur:
* Validate Hypervisor has required iso, network, and resources
* Get the next available vmid from the Hypervisor
* Create the Fluxnode on the Hypervisor
* Start the node (if configured)
"""
        all_dialog = """Add console password and click Provision to configure all fluxnodes.
The following will occur:
* Validate Hypervisor has required iso, network, and resources
* Get the next available vmid from the Hypervisor for each node
* Create the Fluxnodes on the Hypervisor
* Start each node syncronously (if configured)
* Wait an optional delay (in seconds) between each Fluxnode start"""

        dialog = all_dialog if self.all else single_dialog

        with container:
            yield Markdown(dialog)
            with Horizontal(id="console-container"):
                yield Label("Console Pass:", classes="text-label")
                yield Input(
                    placeholder="Enter Console Password",
                    password=True,
                    validators=[PasswordValidator()],
                )
            yield error_label
            with Horizontal(id="button-container"):
                yield Button("Cancel", id="cancel")
                yield Button("Reveal", id="reveal")
                yield Button("Provision", id="provision", disabled=True)

    def on_button_pressed(self, event: Button.Pressed) -> None:
        input = self.query_one(Input)

        if event.button.id == "provision":
            hasher = HashedPassword(input.value)
            self.fluxnode.system.hashed_console = hasher.hash()  # thread?
            self.dismiss(self.fluxnode)
        elif event.button.id == "reveal":
            input.password = not input.password
            event.button.label = "Reveal" if input.password else "Hide"
        else:
            self.dismiss(None)

    def on_input_changed(self, event: Input.Changed) -> None:
        display = bool(event.validation_result.failures)

        label = self.query_one("#error-label", Label)

        label.visible = display

        msg = (
            "\n".join(event.validation_result.failure_descriptions)
            if display
            else ""
        )
        label.update(msg)

        provision_disabled = display or len(event.input.value) < 12

        self.query_one("#provision", Button).disabled = provision_disabled
